@@ -3,7 +3,7 @@
  * Copyright (c) 2004-2006 The Trustees of Indiana University and Indiana
  *                         University Research and Technology
  *                         Corporation.  All rights reserved.
- * Copyright (c) 2004-2009 The University of Tennessee and The University
+ * Copyright (c) 2004-2019 The University of Tennessee and The University
  *                         of Tennessee Research Foundation.  All rights
  *                         reserved.
  * Copyright (c) 2004-2006 High Performance Computing Center Stuttgart,
@@ -11,7 +11,9 @@
  * Copyright (c) 2004-2006 The Regents of the University of California.
  *                         All rights reserved.
  * Copyright (c) 2009      Oak Ridge National Labs.  All rights reserved.
- * Copyright (c) 2014 Cisco Systems, Inc.  All rights reserved.
+ * Copyright (c) 2014      Cisco Systems, Inc.  All rights reserved.
+ * Copyright (c) 2017      Research Organization for Information Science
+ *                         and Technology (RIST). All rights reserved.
  * $COPYRIGHT$
  *
  * Additional copyrights may follow
@@ -34,19 +36,19 @@
 #define UNSET_CONTIGUOUS_FLAG( INT_VALUE )   (INT_VALUE) = (INT_VALUE) & (~(OPAL_DATATYPE_FLAG_CONTIGUOUS | OPAL_DATATYPE_FLAG_NO_GAPS))
 
 #if defined(__GNUC__) && !defined(__STDC__)
-#define LMAX(A,B)  ({ OPAL_PTRDIFF_TYPE _a = (A), _b = (B); (_a < _b ? _b : _a) })
-#define LMIN(A,B)  ({ OPAL_PTRDIFF_TYPE _a = (A), _b = (B); (_a < _b ? _a : _b); })
+#define LMAX(A,B)  ({ ptrdiff_t _a = (A), _b = (B); (_a < _b ? _b : _a) })
+#define LMIN(A,B)  ({ ptrdiff_t _a = (A), _b = (B); (_a < _b ? _a : _b); })
 #define IMAX(A,B)  ({ int _a = (A), _b = (B); (_a < _b ? _b : _a); })
 #else
-static inline OPAL_PTRDIFF_TYPE LMAX( OPAL_PTRDIFF_TYPE a, OPAL_PTRDIFF_TYPE b ) { return ( a < b ? b : a ); }
-static inline OPAL_PTRDIFF_TYPE LMIN( OPAL_PTRDIFF_TYPE a, OPAL_PTRDIFF_TYPE b ) { return ( a < b ? a : b ); }
+static inline ptrdiff_t LMAX( ptrdiff_t a, ptrdiff_t b ) { return ( a < b ? b : a ); }
+static inline ptrdiff_t LMIN( ptrdiff_t a, ptrdiff_t b ) { return ( a < b ? a : b ); }
 static inline int  IMAX( int a, int b ) { return ( a < b ? b : a ); }
 #endif  /* __GNU__ */
 
 #define OPAL_DATATYPE_COMPUTE_REQUIRED_ENTRIES( _pdtAdd, _count, _extent, _place_needed) \
 { \
     if( (_pdtAdd)->flags & OPAL_DATATYPE_FLAG_PREDEFINED ) { /* add a basic datatype */ \
-        (_place_needed) = ((_extent) == (OPAL_PTRDIFF_TYPE)(_pdtAdd)->size ? 1 : 3); \
+        (_place_needed) = ((_extent) == (ptrdiff_t)(_pdtAdd)->size ? 1 : 3); \
     } else { \
         (_place_needed) = (_pdtAdd)->desc.used; \
         if( (_count) != 1 ) { \
@@ -70,7 +72,7 @@ static inline int  IMAX( int a, int b ) { return ( a < b ? b : a ); }
         _new_lb = (_old_lb) + (_disp); \
         _new_ub = (_old_ub) + (_disp); \
     } else { \
-        OPAL_PTRDIFF_TYPE lower, upper; \
+        ptrdiff_t lower, upper; \
         upper = (_disp) + (_old_extent) * ((_count) - 1); \
         lower = (_disp); \
         if( lower < upper ) { \
@@ -101,24 +103,23 @@ static inline int  IMAX( int a, int b ) { return ( a < b ? b : a ); }
  * set to ZERO if it's a empty datatype.
  */
 int32_t opal_datatype_add( opal_datatype_t* pdtBase, const opal_datatype_t* pdtAdd,
-                           uint32_t count, OPAL_PTRDIFF_TYPE disp, OPAL_PTRDIFF_TYPE extent )
+                           size_t count, ptrdiff_t disp, ptrdiff_t extent )
 {
     uint32_t newLength, place_needed = 0, i;
     short localFlags = 0;  /* no specific options yet */
     dt_elem_desc_t *pLast, *pLoop = NULL;
-    OPAL_PTRDIFF_TYPE lb, ub, true_lb, true_ub, epsilon, old_true_ub;
+    ptrdiff_t lb, ub, true_lb, true_ub, epsilon, old_true_ub;
 
     /**
-     *from MPI-3, page 84, lines 18-20: Most datatype constructors have
+     * From MPI-3, page 84, lines 18-20: Most datatype constructors have
      * replication count or block length arguments. Allowed values are
      * non-negative integers. If the value is zero, no elements are generated in
      * the type map and there is no effect on datatype bounds or extent.
      */
     if( 0 == count ) return OPAL_SUCCESS;
 
-    /* the extent should always be positive. So a negative
-     * value here have a special meaning ie. default extent as
-     * computed by ub - lb
+    /* the extent should always be positive. So a negative value here have a
+     * special meaning ie. default extent as computed by ub - lb
      */
     if( extent == -1 ) extent = (pdtAdd->ub - pdtAdd->lb);
 
@@ -131,7 +132,7 @@ int32_t opal_datatype_add( opal_datatype_t* pdtBase, const opal_datatype_t* pdtA
             pdtBase->lb = disp;
             pdtBase->flags |= OPAL_DATATYPE_FLAG_USER_LB;
         }
-        if( (pdtBase->ub - pdtBase->lb) != (OPAL_PTRDIFF_TYPE)pdtBase->size ) {
+        if( (pdtBase->ub - pdtBase->lb) != (ptrdiff_t)pdtBase->size ) {
             pdtBase->flags &= ~OPAL_DATATYPE_FLAG_NO_GAPS;
         }
         return OPAL_SUCCESS; /* Just ignore the OPAL_DATATYPE_LOOP and OPAL_DATATYPE_END_LOOP */
@@ -143,7 +144,7 @@ int32_t opal_datatype_add( opal_datatype_t* pdtBase, const opal_datatype_t* pdtA
             pdtBase->ub = disp;
             pdtBase->flags |= OPAL_DATATYPE_FLAG_USER_UB;
         }
-        if( (pdtBase->ub - pdtBase->lb) != (OPAL_PTRDIFF_TYPE)pdtBase->size ) {
+        if( (pdtBase->ub - pdtBase->lb) != (ptrdiff_t)pdtBase->size ) {
             pdtBase->flags &= ~OPAL_DATATYPE_FLAG_NO_GAPS;
         }
         return OPAL_SUCCESS; /* Just ignore the OPAL_DATATYPE_LOOP and OPAL_DATATYPE_END_LOOP */
@@ -278,45 +279,68 @@ int32_t opal_datatype_add( opal_datatype_t* pdtBase, const opal_datatype_t* pdtA
      * predefined non contiguous datatypes (like MPI_SHORT_INT).
      */
     if( (pdtAdd->flags & (OPAL_DATATYPE_FLAG_PREDEFINED | OPAL_DATATYPE_FLAG_DATA)) == (OPAL_DATATYPE_FLAG_PREDEFINED | OPAL_DATATYPE_FLAG_DATA) ) {
-        pdtBase->btypes[pdtAdd->id] += count;
-        if( (extent != (OPAL_PTRDIFF_TYPE)pdtAdd->size) && (count > 1) ) {  /* gaps around the datatype */
-            localFlags = pdtAdd->flags & ~(OPAL_DATATYPE_FLAG_COMMITTED | OPAL_DATATYPE_FLAG_CONTIGUOUS | OPAL_DATATYPE_FLAG_NO_GAPS);
-            CREATE_LOOP_START( pLast, count, 2, extent, localFlags );
-            pLast++;
-            pLast->elem.common.type  = pdtAdd->id;
-            pLast->elem.count        = 1;
-            pLast->elem.disp         = disp;
-            pLast->elem.extent       = pdtAdd->size;
-            pLast->elem.common.flags = localFlags | OPAL_DATATYPE_FLAG_CONTIGUOUS;
-            pLast++;
-            CREATE_LOOP_END( pLast, 2, disp, pdtAdd->size, localFlags );
-            pdtBase->desc.used += 3;
-            pdtBase->btypes[OPAL_DATATYPE_LOOP]     = 1;
-            pdtBase->btypes[OPAL_DATATYPE_END_LOOP] = 1;
-        } else {
-            pLast->elem.common.type = pdtAdd->id;
-            pLast->elem.count       = count;
-            pLast->elem.disp        = disp;
-            pLast->elem.extent      = extent;
-            pdtBase->desc.used++;
-            pLast->elem.common.flags  = pdtAdd->flags & ~(OPAL_DATATYPE_FLAG_COMMITTED);
+        if( NULL != pdtBase->ptypes )
+            pdtBase->ptypes[pdtAdd->id] += count;
+
+        pLast->elem.common.flags     = pdtAdd->flags & ~(OPAL_DATATYPE_FLAG_COMMITTED);
+        pLast->elem.common.type      = pdtAdd->id;
+        pLast->elem.disp             = disp;
+        pLast->elem.extent           = count * extent;
+        /* assume predefined datatypes without extent, aka. contiguous */
+        pLast->elem.count            = 1;
+        pLast->elem.blocklen         = count;
+        if( extent != (ptrdiff_t)pdtAdd->size ) {  /* not contiguous: let's fix */
+            pLast->elem.count            = count;
+            pLast->elem.blocklen         = 1;
+            pLast->elem.extent           = extent;
+            if( count > 1 ) {  /* gaps around the predefined datatype */
+                pLast->elem.common.flags &= ~(OPAL_DATATYPE_FLAG_CONTIGUOUS | OPAL_DATATYPE_FLAG_NO_GAPS);
+            }
         }
+        pdtBase->desc.used++;
     } else {
         /* keep trace of the total number of basic datatypes in the datatype definition */
-        pdtBase->btypes[OPAL_DATATYPE_LOOP]     += pdtAdd->btypes[OPAL_DATATYPE_LOOP];
-        pdtBase->btypes[OPAL_DATATYPE_END_LOOP] += pdtAdd->btypes[OPAL_DATATYPE_END_LOOP];
-        pdtBase->btypes[OPAL_DATATYPE_LB]       |= pdtAdd->btypes[OPAL_DATATYPE_LB];
-        pdtBase->btypes[OPAL_DATATYPE_UB]       |= pdtAdd->btypes[OPAL_DATATYPE_UB];
-        for( i = 4; i < OPAL_DATATYPE_MAX_PREDEFINED; i++ )
-            if( pdtAdd->btypes[i] != 0 ) pdtBase->btypes[i] += (count * pdtAdd->btypes[i]);
-
-        if( (1 == pdtAdd->desc.used) && (extent == (pdtAdd->ub - pdtAdd->lb)) &&
-            (extent == pdtAdd->desc.desc[0].elem.extent) ){
+        pdtBase->loops += pdtAdd->loops;
+        pdtBase->flags |= (pdtAdd->flags & OPAL_DATATYPE_FLAG_USER_LB);
+        pdtBase->flags |= (pdtAdd->flags & OPAL_DATATYPE_FLAG_USER_UB);
+        if( (NULL != pdtBase->ptypes) && (NULL != pdtAdd->ptypes) ) {
+            for( i = OPAL_DATATYPE_FIRST_TYPE; i < OPAL_DATATYPE_MAX_PREDEFINED; i++ )
+                if( pdtAdd->ptypes[i] != 0 ) pdtBase->ptypes[i] += (count * pdtAdd->ptypes[i]);
+        }
+        if( 1 == pdtAdd->desc.used ) {
             pLast->elem        = pdtAdd->desc.desc[0].elem;
-            pLast->elem.count *= count;
             pLast->elem.disp  += disp;
+            if( 1 == count ) {
+                /* Extent only has a meaning when there are multiple elements. Bail out */
+            } else if( 1 == pLast->elem.count ) {
+                /* The size and true_extent of the added datatype are identical, signaling a datatype
+                 * that is mostly contiguous with the exception of the initial and final gaps. These
+                 * gaps do not matter here as they will amended (the initial gaps being shifted by the
+                 * new displacement and the final gap being replaced with the new gap
+                 */
+                if( pdtAdd->desc.desc[0].elem.extent == extent ) {
+                    /* pure bliss everything is fully contiguous and we can collapse
+                     * everything by updating the blocklen and extent
+                     */
+                    pLast->elem.blocklen *= count;
+                    pLast->elem.extent   *= count;
+                } else {
+                    pLast->elem.count = count;
+                    pLast->elem.extent = extent;
+                }
+            } else if( extent == (ptrdiff_t)(pLast->elem.count * pLast->elem.extent) ) {
+                /* It's just a repetition of the same element, increase the count */
+                pLast->elem.count *= count;
+            } else {
+                /* No luck here, no optimization can be applied. Fall back to the
+                 * normal case where we add a loop around the datatype.
+                 */
+                goto build_loop;
+            }
             pdtBase->desc.used++;
         } else {
+
+build_loop:
             /* if the extent of the datatype is the same as the extent of the loop
              * description of the datatype then we simply have to update the main loop.
              */
@@ -324,7 +348,7 @@ int32_t opal_datatype_add( opal_datatype_t* pdtBase, const opal_datatype_t* pdtA
                 pLoop = pLast;
                 CREATE_LOOP_START( pLast, count, pdtAdd->desc.used + 1, extent,
                                    (pdtAdd->flags & ~(OPAL_DATATYPE_FLAG_COMMITTED)) );
-                pdtBase->btypes[OPAL_DATATYPE_LOOP] += 2;
+                pdtBase->loops     += 2;
                 pdtBase->desc.used += 2;
                 pLast++;
             }
@@ -358,11 +382,11 @@ int32_t opal_datatype_add( opal_datatype_t* pdtBase, const opal_datatype_t* pdtA
     UNSET_CONTIGUOUS_FLAG(pdtBase->flags);
     if( (localFlags & OPAL_DATATYPE_FLAG_CONTIGUOUS)             /* both type were contiguous */
         && ((disp + pdtAdd->true_lb) == old_true_ub)  /* and there is no gap between them */
-        && ( ((OPAL_PTRDIFF_TYPE)pdtAdd->size == extent)      /* the size and the extent of the
+        && ( ((ptrdiff_t)pdtAdd->size == extent)      /* the size and the extent of the
                                                        * added type have to match */
              || (count < 2)) ) {                      /* if the count is bigger than 2 */
             SET_CONTIGUOUS_FLAG(pdtBase->flags);
-            if( (OPAL_PTRDIFF_TYPE)pdtBase->size == (pdtBase->ub - pdtBase->lb) )
+            if( (ptrdiff_t)pdtBase->size == (pdtBase->ub - pdtBase->lb) )
                 SET_NO_GAP_FLAG(pdtBase->flags);
     }
 
