@@ -13,9 +13,9 @@
  * Copyright (c) 2007-2012 Los Alamos National Security, LLC.  All rights
  *                         reserved.
  * Copyright (c) 2008-2009 Sun Microsystems, Inc.  All rights reserved.
- * Copyright (c) 2011      IBM Corporation.  All rights reserved.
- * Copyright (c) 2014-2015 Intel Corporation.  All rights reserved.
- * Copyright (c) 2015      Research Organization for Information Science
+ * Copyright (c) 2011-2017 IBM Corporation.  All rights reserved.
+ * Copyright (c) 2014-2018 Intel, Inc. All rights reserved.
+ * Copyright (c) 2015-2018 Research Organization for Information Science
  *                         and Technology (RIST). All rights reserved.
  * $COPYRIGHT$
  *
@@ -78,14 +78,16 @@
 #include "orte/runtime/orte_wait.h"
 #include "orte/runtime/orte_globals.h"
 #include "orte/util/name_fns.h"
-#include "orte/util/nidmap.h"
 #include "orte/util/proc_info.h"
+#include "orte/util/threads.h"
 
 #include "orte/mca/rml/rml.h"
 #include "orte/mca/rml/rml_types.h"
 #include "orte/mca/ess/ess.h"
 #include "orte/mca/ess/base/base.h"
 #include "orte/mca/errmgr/errmgr.h"
+#include "orte/mca/grpcomm/base/base.h"
+#include "orte/mca/oob/base/base.h"
 #include "orte/mca/rmaps/rmaps.h"
 #include "orte/mca/routed/routed.h"
 #include "orte/mca/rml/base/rml_contact.h"
@@ -98,7 +100,7 @@
 
 static int rsh_init(void);
 static int rsh_launch(orte_job_t *jdata);
-static int remote_spawn(opal_buffer_t *launch);
+static int remote_spawn(void);
 static int rsh_terminate_orteds(void);
 static int rsh_finalize(void);
 
@@ -164,6 +166,7 @@ static const char *orte_plm_rsh_shell_name[7] = {
  */
 static void set_handler_default(int sig);
 static orte_plm_rsh_shell_t find_shell(char *shell);
+static int launch_agent_setup(const char *agent, char *path);
 static void ssh_child(int argc, char **argv) __opal_attribute_noreturn__;
 static int rsh_probe(char *nodename,
                      orte_plm_rsh_shell_t *shell);
@@ -177,14 +180,59 @@ static void process_launch_list(int fd, short args, void *cbdata);
 static int num_in_progress=0;
 static opal_list_t launch_list;
 static opal_event_t launch_event;
+static char *rsh_agent_path=NULL;
+static char **rsh_agent_argv=NULL;
 
 /**
  * Init the module
  */
 static int rsh_init(void)
 {
+    char *tmp;
     int rc;
 
+    /* we were selected, so setup the launch agent */
+    if (mca_plm_rsh_component.using_qrsh) {
+        /* perform base setup for qrsh */
+        asprintf(&tmp, "%s/bin/%s", getenv("SGE_ROOT"), getenv("ARC"));
+        if (ORTE_SUCCESS != (rc = launch_agent_setup("qrsh", tmp))) {
+            ORTE_ERROR_LOG(rc);
+            free(tmp);
+            return rc;
+        }
+        free(tmp);
+        /* automatically add -inherit and grid engine PE related flags */
+        opal_argv_append_nosize(&rsh_agent_argv, "-inherit");
+        /* Don't use the "-noshell" flag as qrsh would have a problem
+         * swallowing a long command */
+        opal_argv_append_nosize(&rsh_agent_argv, "-nostdin");
+        opal_argv_append_nosize(&rsh_agent_argv, "-V");
+        if (0 < opal_output_get_verbosity(orte_plm_base_framework.framework_output)) {
+            opal_argv_append_nosize(&rsh_agent_argv, "-verbose");
+            tmp = opal_argv_join(rsh_agent_argv, ' ');
+            opal_output_verbose(1, orte_plm_base_framework.framework_output,
+                                "%s plm:rsh: using \"%s\" for launching\n",
+                                ORTE_NAME_PRINT(ORTE_PROC_MY_NAME), tmp);
+            free(tmp);
+        }
+    } else if(mca_plm_rsh_component.using_llspawn) {
+        /* perform base setup for llspawn */
+        if (ORTE_SUCCESS != (rc = launch_agent_setup("llspawn", NULL))) {
+            ORTE_ERROR_LOG(rc);
+            return rc;
+        }
+        opal_output_verbose(1, orte_plm_base_framework.framework_output,
+                            "%s plm:rsh: using \"%s\" for launching\n",
+                            ORTE_NAME_PRINT(ORTE_PROC_MY_NAME),
+                            rsh_agent_path);
+    } else {
+        /* not using qrsh or llspawn - use MCA-specified agent */
+        if (ORTE_SUCCESS != (rc = launch_agent_setup(mca_plm_rsh_component.agent, NULL))) {
+            ORTE_ERROR_LOG(rc);
+            return rc;
+        }
+    }
+
     /* point to our launch command */
     if (ORTE_SUCCESS != (rc = orte_state.add_job_state(ORTE_JOB_STATE_LAUNCH_DAEMONS,
                                                        launch_daemons, ORTE_SYS_PRI))) {
@@ -211,21 +259,25 @@ static int rsh_init(void)
 /**
  * Callback on daemon exit.
  */
-static void rsh_wait_daemon(orte_proc_t *daemon, void* cbdata)
+static void rsh_wait_daemon(int sd, short flags, void *cbdata)
 {
     orte_job_t *jdata;
-    orte_plm_rsh_caddy_t *caddy=(orte_plm_rsh_caddy_t*)cbdata;
+    orte_wait_tracker_t *t2 = (orte_wait_tracker_t*)cbdata;
+    orte_plm_rsh_caddy_t *caddy=(orte_plm_rsh_caddy_t*)t2->cbdata;
+    orte_proc_t *daemon = caddy->daemon;
+    char *rtmod;
 
     if (orte_orteds_term_ordered || orte_abnormal_term_ordered) {
         /* ignore any such report - it will occur if we left the
          * session attached, e.g., while debugging
          */
         OBJ_RELEASE(caddy);
+        OBJ_RELEASE(t2);
         return;
     }
 
-    if (! WIFEXITED(daemon->exit_code) ||
-        ! WEXITSTATUS(daemon->exit_code) == 0) { /* if abnormal exit */
+    if (!WIFEXITED(daemon->exit_code) ||
+        WEXITSTATUS(daemon->exit_code) != 0) { /* if abnormal exit */
         /* if we are not the HNP, send a message to the HNP alerting it
          * to the failure
          */
@@ -238,7 +290,8 @@ static void rsh_wait_daemon(orte_proc_t *daemon, void* cbdata)
             buf = OBJ_NEW(opal_buffer_t);
             opal_dss.pack(buf, &(daemon->name.vpid), 1, ORTE_VPID);
             opal_dss.pack(buf, &daemon->exit_code, 1, OPAL_INT);
-            orte_rml.send_buffer_nb(ORTE_PROC_MY_HNP, buf,
+            orte_rml.send_buffer_nb(orte_coll_conduit,
+                                    ORTE_PROC_MY_HNP, buf,
                                     ORTE_RML_TAG_REPORT_REMOTE_LAUNCH,
                                     orte_rml_send_callback, NULL);
             /* note that this daemon failed */
@@ -259,7 +312,8 @@ static void rsh_wait_daemon(orte_proc_t *daemon, void* cbdata)
             /* remove it from the routing table to ensure num_routes
              * returns the correct value
              */
-            orte_routed.route_lost(&daemon->name);
+            rtmod = orte_rml.get_routed(orte_coll_conduit);
+            orte_routed.route_lost(rtmod, &daemon->name);
             /* report that the daemon has failed so we can exit */
             ORTE_ACTIVATE_PROC_STATE(&daemon->name, ORTE_PROC_STATE_FAILED_TO_START);
         }
@@ -272,7 +326,7 @@ static void rsh_wait_daemon(orte_proc_t *daemon, void* cbdata)
         opal_event_active(&launch_event, EV_WRITE, 1);
     }
     /* cleanup */
-    OBJ_RELEASE(caddy);
+    OBJ_RELEASE(t2);
 }
 
 static int setup_launch(int *argcptr, char ***argvptr,
@@ -289,11 +343,12 @@ static int setup_launch(int *argcptr, char ***argvptr,
     char *orted_cmd, *orted_prefix, *final_cmd;
     int orted_index;
     int rc;
-    int i, j;
+    int i, j, cnt;
     bool found;
     char *lib_base=NULL, *bin_base=NULL;
     char *opal_prefix = getenv("OPAL_PREFIX");
     char* full_orted_cmd = NULL;
+    char * rtmod;
 
     /* Figure out the basenames for the libdir and bindir.  This
        requires some explanation:
@@ -325,8 +380,8 @@ static int setup_launch(int *argcptr, char ***argvptr,
     /*
      * Build argv array
      */
-    argv = opal_argv_copy(mca_plm_rsh_component.agent_argv);
-    argc = opal_argv_count(mca_plm_rsh_component.agent_argv);
+    argv = opal_argv_copy(rsh_agent_argv);
+    argc = opal_argv_count(argv);
     /* if any ssh args were provided, now is the time to add them */
     if (NULL != mca_plm_rsh_component.ssh_args) {
         char **ssh_argv;
@@ -553,29 +608,53 @@ static int setup_launch(int *argcptr, char ***argvptr,
          (mca_plm_rsh_component.using_qrsh && mca_plm_rsh_component.daemonize_qrsh)) &&
         ((!mca_plm_rsh_component.using_llspawn) ||
          (mca_plm_rsh_component.using_llspawn && mca_plm_rsh_component.daemonize_llspawn))) {
-        opal_argv_append(&argc, &argv, "--daemonize");
+    }
+
+    if (!mca_plm_rsh_component.no_tree_spawn) {
+        // Remove problematic and/or conflicting command line arguments that
+        // should not be passed on to our children.
+        cnt = opal_argv_count(orted_cmd_line);
+        for (i=0; i < cnt; i+=3) {
+            if (0 == strcmp(orted_cmd_line[i+1], "routed")) {
+                opal_argv_delete(&cnt, &orted_cmd_line, i, 3);
+                break;
+            }
+        }
     }
 
     /*
      * Add the basic arguments to the orted command line, including
      * all debug options
      */
-    if (ORTE_PROC_IS_CM) {
-        orte_plm_base_orted_append_basic_args(&argc, &argv,
-                                              NULL,
-                                              proc_vpid_index,
-                                              NULL);
-    } else {
-        orte_plm_base_orted_append_basic_args(&argc, &argv,
-                                              "env",
-                                              proc_vpid_index,
-                                              NULL);
-    }
+    orte_plm_base_orted_append_basic_args(&argc, &argv,
+                                          "env",
+                                          proc_vpid_index);
 
     /* ensure that only the ssh plm is selected on the remote daemon */
-    opal_argv_append_nosize(&argv, "-"OPAL_MCA_CMD_LINE_ID);
-    opal_argv_append_nosize(&argv, "plm");
-    opal_argv_append_nosize(&argv, "rsh");
+    opal_argv_append(&argc, &argv, "-"OPAL_MCA_CMD_LINE_ID);
+    opal_argv_append(&argc, &argv, "plm");
+    opal_argv_append(&argc, &argv, "rsh");
+
+    /* if we are tree-spawning, tell our child daemons the
+     * uri of their parent (me) */
+    if (!mca_plm_rsh_component.no_tree_spawn) {
+        opal_argv_append(&argc, &argv, "--tree-spawn");
+        orte_oob_base_get_addr(&param);
+
+        // When tree-spawn'ing we need to force the remote daemons to use
+        // the routing component that was used to setup the launch tree.
+        // Otherwise the orte_parent_uri will not match the orted they
+        // expect to find in the routing tree.
+        rtmod = orte_rml.get_routed(orte_coll_conduit);
+        opal_argv_append(&argc, &argv, "-"OPAL_MCA_CMD_LINE_ID);
+        opal_argv_append(&argc, &argv, "routed");
+        opal_argv_append(&argc, &argv, rtmod);
+
+        opal_argv_append(&argc, &argv, "-"OPAL_MCA_CMD_LINE_ID);
+        opal_argv_append(&argc, &argv, "orte_parent_uri");
+        opal_argv_append(&argc, &argv, param);
+        free(param);
+    }
 
     /* unless told otherwise... */
     if (mca_plm_rsh_component.pass_environ_mca_params) {
@@ -676,7 +755,7 @@ static void ssh_child(int argc, char **argv)
      * about remote launches here
      */
     exec_argv = argv;
-    exec_path = mca_plm_rsh_component.agent_path;
+    exec_path = strdup(rsh_agent_path);
 
     /* Don't let ssh slurp all of our stdin! */
     fdin = open("/dev/null", O_RDWR);
@@ -727,7 +806,7 @@ static void ssh_child(int argc, char **argv)
 /*
  * launch a set of daemons from a remote daemon
  */
-static int remote_spawn(opal_buffer_t *launch)
+static int remote_spawn(void)
 {
     int node_name_index1;
     int proc_vpid_index;
@@ -736,13 +815,12 @@ static int remote_spawn(opal_buffer_t *launch)
     int argc;
     int rc=ORTE_SUCCESS;
     bool failed_launch = true;
-    orte_std_cntr_t n;
-    opal_byte_object_t *bo;
     orte_process_name_t target;
     orte_plm_rsh_caddy_t *caddy;
     orte_job_t *daemons;
     opal_list_t coll;
     orte_namelist_t *child;
+    char *rtmod;
 
     OPAL_OUTPUT_VERBOSE((1, orte_plm_base_framework.framework_output,
                          "%s plm:rsh: remote spawn called",
@@ -751,31 +829,19 @@ static int remote_spawn(opal_buffer_t *launch)
     /* if we hit any errors, tell the HNP it was us */
     target.vpid = ORTE_PROC_MY_NAME->vpid;
 
-    /* extract the prefix from the launch buffer */
-    n = 1;
-    if (ORTE_SUCCESS != (rc = opal_dss.unpack(launch, &prefix, &n, OPAL_STRING))) {
-        ORTE_ERROR_LOG(rc);
-        goto cleanup;
+    /* check to see if enable-orterun-prefix-by-default was given - if
+     * this is being done by a singleton, then orterun will not be there
+     * to put the prefix in the app. So make sure we check to find it */
+    if ((bool)ORTE_WANT_ORTERUN_PREFIX_BY_DEFAULT) {
+        prefix = strdup(opal_install_dirs.prefix);
+    } else {
+        prefix = NULL;
     }
 
-    /* extract the byte object holding the nidmap */
-    n=1;
-    if (ORTE_SUCCESS != (rc = opal_dss.unpack(launch, &bo, &n, OPAL_BYTE_OBJECT))) {
-        ORTE_ERROR_LOG(rc);
-        goto cleanup;
-    }
-    /* update our nidmap - this will free data in the byte object */
-    if (ORTE_SUCCESS != (rc = orte_util_decode_daemon_nodemap(bo))) {
-        ORTE_ERROR_LOG(rc);
-        goto cleanup;
-    }
-
-    /* ensure the routing plan is updated */
-    orte_routed.update_routing_plan();
-
     /* get the updated routing list */
+    rtmod = orte_rml.get_routed(orte_coll_conduit);
     OBJ_CONSTRUCT(&coll, opal_list_t);
-    orte_routed.get_routing_list(&coll);
+    orte_routed.get_routing_list(rtmod, &coll);
 
     /* if I have no children, just return */
     if (0 == opal_list_get_size(&coll)) {
@@ -789,7 +855,8 @@ static int remote_spawn(opal_buffer_t *launch)
     }
 
     /* setup the launch */
-    if (ORTE_SUCCESS != (rc = setup_launch(&argc, &argv, orte_process_info.nodename, &node_name_index1,
+    if (ORTE_SUCCESS != (rc = setup_launch(&argc, &argv,
+                                           orte_process_info.nodename, &node_name_index1,
                                            &proc_vpid_index, prefix))) {
         ORTE_ERROR_LOG(rc);
         OBJ_DESTRUCT(&coll);
@@ -843,6 +910,10 @@ static int remote_spawn(opal_buffer_t *launch)
         opal_list_append(&launch_list, &caddy->super);
     }
     OPAL_LIST_DESTRUCT(&coll);
+    /* we NEVER use tree-spawn for secondary launches - e.g.,
+     * due to a dynamic launch requesting add_hosts - so be
+     * sure to turn it off here */
+    mca_plm_rsh_component.no_tree_spawn = true;
 
     /* trigger the event to start processing the launch list */
     OPAL_OUTPUT_VERBOSE((1, orte_plm_base_framework.framework_output,
@@ -865,7 +936,8 @@ cleanup:
         buf = OBJ_NEW(opal_buffer_t);
         opal_dss.pack(buf, &target.vpid, 1, ORTE_VPID);
         opal_dss.pack(buf, &rc, 1, OPAL_INT);
-        orte_rml.send_buffer_nb(ORTE_PROC_MY_HNP, buf,
+        orte_rml.send_buffer_nb(orte_coll_conduit,
+                                ORTE_PROC_MY_HNP, buf,
                                 ORTE_RML_TAG_REPORT_REMOTE_LAUNCH,
                                 orte_rml_send_callback, NULL);
     }
@@ -896,6 +968,8 @@ static void process_launch_list(int fd, short args, void *cbdata)
     pid_t pid;
     orte_plm_rsh_caddy_t *caddy;
 
+    ORTE_ACQUIRE_OBJECT(caddy);
+
     while (num_in_progress < mca_plm_rsh_component.num_concurrent) {
         item = opal_list_remove_first(&launch_list);
         if (NULL == item) {
@@ -905,7 +979,7 @@ static void process_launch_list(int fd, short args, void *cbdata)
         caddy = (orte_plm_rsh_caddy_t*)item;
         /* register the sigchild callback */
         ORTE_FLAG_SET(caddy->daemon, ORTE_PROC_FLAG_ALIVE);
-        orte_wait_cb(caddy->daemon, rsh_wait_daemon, (void*)caddy);
+        orte_wait_cb(caddy->daemon, rsh_wait_daemon, orte_event_base, (void*)caddy);
 
         /* fork a child to exec the rsh/ssh session */
         pid = fork();
@@ -917,9 +991,45 @@ static void process_launch_list(int fd, short args, void *cbdata)
 
         /* child */
         if (pid == 0) {
+            /*
+             * When the user presses CTRL-C, SIGINT is sent to the whole process
+             * group which terminates the rsh/ssh command. This can cause the
+             * remote daemon to crash with a SIGPIPE when it tried to print out
+             * status information. This has two concequences:
+             * 1) The remote node is not cleaned up as it should. The local
+             *    processes will notice that the orted failed and cleanup their
+             *    part of the session directory, but the job level part will
+             *    remain littered.
+             * 2) Any debugging information we expected to see from the orted
+             *    during shutdown is lost.
+             *
+             * The solution here is to put the child processes in a separate
+             * process group from the HNP. So when the user presses CTRL-C
+             * then only the HNP receives the signal, and not the rsh/ssh
+             * child processes.
+             */
+#if HAVE_SETPGID
+            if( 0 != setpgid(0, 0) ) {
+                opal_output(0, "plm:rsh: Error: setpgid(0,0) failed in child with errno=%s(%d)\n",
+                            strerror(errno), errno);
+                exit(-1);
+            }
+#endif
+
             /* do the ssh launch - this will exit if it fails */
             ssh_child(caddy->argc, caddy->argv);
         } else { /* father */
+            // Put the child in a separate progress group
+            // - see comment in child section.
+#if HAVE_SETPGID
+            if( 0 != setpgid(pid, pid) ) {
+                opal_output(0, "plm:rsh: Warning: setpgid(%ld,%ld) failed in parent with errno=%s(%d)\n",
+                            (long)pid, (long)pid, strerror(errno), errno);
+                // Ignore this error since the child is off and running.
+                // We still need to track it.
+            }
+#endif
+
             /* indicate this daemon has been launched */
             caddy->daemon->state = ORTE_PROC_STATE_RUNNING;
             /* record the pid of the ssh fork */
@@ -951,7 +1061,11 @@ static void launch_daemons(int fd, short args, void *cbdata)
     orte_plm_rsh_caddy_t *caddy;
     opal_list_t coll;
     char *username;
+    int port, *portptr;
     orte_namelist_t *child;
+    char *rtmod;
+
+    ORTE_ACQUIRE_OBJECT(state);
 
     /* if we are launching debugger daemons, then just go
      * do it - no new daemons will be launched
@@ -1027,7 +1141,8 @@ static void launch_daemons(int fd, short args, void *cbdata)
                        true, mca_plm_rsh_component.num_concurrent, map->num_new_daemons);
         ORTE_ERROR_LOG(ORTE_ERR_FATAL);
         OBJ_RELEASE(state);
-        return;
+        rc = ORTE_ERR_SILENT;
+        goto cleanup;
     }
 
     /*
@@ -1072,43 +1187,18 @@ static void launch_daemons(int fd, short args, void *cbdata)
             }
         }
     }
+    if (NULL == node) {
+        /* this should be impossible, but adding the check will
+         * silence code checkers that don't know better */
+        ORTE_ERROR_LOG(ORTE_ERR_NOT_FOUND);
+        rc = ORTE_ERR_NOT_FOUND;
+        goto cleanup;
+    }
 
     /* if we are tree launching, find our children and create the launch cmd */
     if (!mca_plm_rsh_component.no_tree_spawn) {
-        orte_daemon_cmd_flag_t command = ORTE_DAEMON_TREE_SPAWN;
-        opal_byte_object_t bo, *boptr;
         orte_job_t *jdatorted;
 
-        /* get the tree spawn buffer */
-        orte_tree_launch_cmd = OBJ_NEW(opal_buffer_t);
-        /* insert the tree_spawn cmd */
-        if (ORTE_SUCCESS != (rc = opal_dss.pack(orte_tree_launch_cmd, &command, 1, ORTE_DAEMON_CMD))) {
-            ORTE_ERROR_LOG(rc);
-            OBJ_RELEASE(orte_tree_launch_cmd);
-            goto cleanup;
-        }
-        /* pack the prefix since this will be needed by the next wave */
-        if (ORTE_SUCCESS != (rc = opal_dss.pack(orte_tree_launch_cmd, &prefix_dir, 1, OPAL_STRING))) {
-            ORTE_ERROR_LOG(rc);
-            OBJ_RELEASE(orte_tree_launch_cmd);
-            goto cleanup;
-        }
-        /* construct a nodemap of all daemons we know about */
-        if (ORTE_SUCCESS != (rc = orte_util_encode_nodemap(&bo, false))) {
-            ORTE_ERROR_LOG(rc);
-            OBJ_RELEASE(orte_tree_launch_cmd);
-            goto cleanup;
-        }
-        /* store it */
-        boptr = &bo;
-        if (ORTE_SUCCESS != (rc = opal_dss.pack(orte_tree_launch_cmd, &boptr, 1, OPAL_BYTE_OBJECT))) {
-            ORTE_ERROR_LOG(rc);
-            OBJ_RELEASE(orte_tree_launch_cmd);
-            free(bo.bytes);
-            goto cleanup;
-        }
-        /* release the data since it has now been copied into our buffer */
-        free(bo.bytes);
         /* get the orted job data object */
         if (NULL == (jdatorted = orte_get_job_data_object(ORTE_PROC_MY_NAME->jobid))) {
             ORTE_ERROR_LOG(ORTE_ERR_NOT_FOUND);
@@ -1118,7 +1208,12 @@ static void launch_daemons(int fd, short args, void *cbdata)
 
         /* get the updated routing list */
         OBJ_CONSTRUCT(&coll, opal_list_t);
-        orte_routed.get_routing_list(&coll);
+        rtmod = orte_rml.get_routed(orte_coll_conduit);
+        orte_routed.get_routing_list(rtmod, &coll);
+
+        OPAL_OUTPUT_VERBOSE((1, orte_plm_base_framework.framework_output,
+                             "%s plm:rsh:launch Tree Launch using routed/%s",
+                             ORTE_NAME_PRINT(ORTE_PROC_MY_NAME), rtmod));
     }
 
     /* setup the launch */
@@ -1203,10 +1298,23 @@ static void launch_daemons(int fd, short args, void *cbdata)
         caddy = OBJ_NEW(orte_plm_rsh_caddy_t);
         caddy->argc = argc;
         caddy->argv = opal_argv_copy(argv);
+        /* insert the alternate port if any */
+        portptr = &port;
+        if (orte_get_attribute(&node->attributes, ORTE_NODE_PORT, (void**)&portptr, OPAL_INT)) {
+            char portname[16];
+            /* for the sake of simplicity, insert "-p" <port> in the duplicated argv */
+            opal_argv_insert_element(&caddy->argv, node_name_index1+1, "-p");
+            snprintf (portname, 15, "%d", port);
+            opal_argv_insert_element(&caddy->argv, node_name_index1+2, portname);
+        }
         caddy->daemon = node->daemon;
         OBJ_RETAIN(caddy->daemon);
         opal_list_append(&launch_list, &caddy->super);
     }
+    /* we NEVER use tree-spawn for secondary launches - e.g.,
+     * due to a dynamic launch requesting add_hosts - so be
+     * sure to turn it off here */
+    mca_plm_rsh_component.no_tree_spawn = true;
 
     /* set the job state to indicate the daemons are launched */
     state->jdata->state = ORTE_JOB_STATE_DAEMONS_LAUNCHED;
@@ -1215,6 +1323,7 @@ static void launch_daemons(int fd, short args, void *cbdata)
     OPAL_OUTPUT_VERBOSE((1, orte_plm_base_framework.framework_output,
                          "%s plm:rsh: activating launch event",
                          ORTE_NAME_PRINT(ORTE_PROC_MY_NAME)));
+    ORTE_POST_OBJECT(state);
     opal_event_active(&launch_event, EV_WRITE, 1);
 
     /* now that we've launched the daemons, let the daemon callback
@@ -1285,6 +1394,10 @@ static int rsh_finalize(void)
             }
         }
     }
+    free(mca_plm_rsh_component.agent_path);
+    free(rsh_agent_path);
+    opal_argv_free(mca_plm_rsh_component.agent_argv);
+    opal_argv_free(rsh_agent_argv);
 
     return rc;
 }
@@ -1331,6 +1444,66 @@ static orte_plm_rsh_shell_t find_shell(char *shell)
     return ORTE_PLM_RSH_SHELL_UNKNOWN;
 }
 
+static int launch_agent_setup(const char *agent, char *path)
+{
+    char *bname;
+    int i;
+
+    /* if no agent was provided, then report not found */
+    if (NULL == mca_plm_rsh_component.agent && NULL == agent) {
+        return ORTE_ERR_NOT_FOUND;
+    }
+
+    /* search for the argv */
+    OPAL_OUTPUT_VERBOSE((5, orte_plm_base_framework.framework_output,
+                         "%s plm:rsh_setup on agent %s path %s",
+                         ORTE_NAME_PRINT(ORTE_PROC_MY_NAME),
+                         (NULL == agent) ? mca_plm_rsh_component.agent : agent,
+                         (NULL == path) ? "NULL" : path));
+    rsh_agent_argv = orte_plm_rsh_search(agent, path);
+
+    if (0 == opal_argv_count(rsh_agent_argv)) {
+        /* nothing was found */
+        return ORTE_ERR_NOT_FOUND;
+    }
+
+    /* see if we can find the agent in the path */
+    rsh_agent_path = opal_path_findv(rsh_agent_argv[0], X_OK, environ, path);
+
+    if (NULL == rsh_agent_path) {
+        /* not an error - just report not found */
+        opal_argv_free(rsh_agent_argv);
+        return ORTE_ERR_NOT_FOUND;
+    }
+
+    bname = opal_basename(rsh_agent_argv[0]);
+    if (NULL != bname && 0 == strcmp(bname, "ssh")) {
+        /* if xterm option was given, add '-X', ensuring we don't do it twice */
+        if (NULL != orte_xterm) {
+            opal_argv_append_unique_nosize(&rsh_agent_argv, "-X", false);
+        } else if (0 >= opal_output_get_verbosity(orte_plm_base_framework.framework_output)) {
+            /* if debug was not specified, and the user didn't explicitly
+             * specify X11 forwarding/non-forwarding, add "-x" if it
+             * isn't already there (check either case)
+             */
+            for (i = 1; NULL != rsh_agent_argv[i]; ++i) {
+                if (0 == strcasecmp("-x", rsh_agent_argv[i])) {
+                    break;
+                }
+            }
+            if (NULL == rsh_agent_argv[i]) {
+                opal_argv_append_nosize(&rsh_agent_argv, "-x");
+            }
+        }
+    }
+    if (NULL != bname) {
+        free(bname);
+    }
+
+    /* the caller can append any additional argv's they desire */
+    return ORTE_SUCCESS;
+}
+
 /**
  * Check the Shell variable and system type on the specified node
  */
@@ -1454,15 +1627,10 @@ static int setup_shell(orte_plm_rsh_shell_t *rshell,
         struct passwd *p;
 
         p = getpwuid(getuid());
-        if( NULL == p ) {
-            /* This user is unknown to the system. Therefore, there is no reason we
-             * spawn whatsoever in his name. Give up with a HUGE error message.
-             */
-            orte_show_help( "help-plm-rsh.txt", "unknown-user", true, (int)getuid() );
-            return ORTE_ERR_FATAL;
+        if( NULL != p ) {
+            param = p->pw_shell;
+            local_shell = find_shell(p->pw_shell);
         }
-        param = p->pw_shell;
-        local_shell = find_shell(p->pw_shell);
     }
 #endif
 
